from .cgp_config import CgpConfig
from .cnn_evolution import CNNEvaluation
from .individual import Individual
from .cgp import CGP

__all__ = [
    'CgpConfig',
    'CNNEvaluation',
    'Individual',
    'CGP'
]
